import os
import traceback
import logging

import hypertrace.agent.filter
from opentelemetry.trace import Span
from traceableai.config.config import Config

logger = logging.getLogger(__name__)

_BLOCKING_AVAILABLE = False

try:
    import traceableai.filter._blocking as _blocking # pylint: disable=import-error,no-name-in-module
    _BLOCKING_AVAILABLE = True
except Exception as error:  # pylint: disable=broad-except
    _BLOCKING_AVAILABLE = False
    traceback.print_tb(error.__traceback__)
    logger.warning(error)
    logger.warning("error loading traceable extension - blocking filter disabled")

HTTP_HEADER_PREFIX = "http.request.header"

class Traceable(hypertrace.agent.filter.Filter):

    def __init__(self):
        if _BLOCKING_AVAILABLE:
            self.blocking_config = _create_blocking_config()
            p_blocking_engine = _blocking.ffi.new("traceable_blocking_engine*")
            result_code = _blocking.lib.traceable_new_blocking_engine(self.blocking_config[0], p_blocking_engine)
            self.blocking_available = (result_code == _blocking.lib.TRACEABLE_SUCCESS)
            if self.blocking_available:
                logger.debug("blocking available")
                self.blocking_engine = p_blocking_engine[0]
                result = _blocking.lib.traceable_start_blocking_engine(self.blocking_engine)
                if result != _blocking.lib.TRACEABLE_SUCCESS:
                    logger.debug("Failed to start blocking engine")
                    self.blocking_available = False
        else:
            self.blocking_available = False

    def evaluate_url_and_headers(self, span: Span, url: str, headers: dict) -> bool:
        if not self.blocking_available:
            return False

        pairs = []

        ip_address = span._readable_span().attributes.get('net.peer.ip') # pylint:disable=W0212
        if ip_address:
            pairs.append((_blocking.ffi.new("char[]", b"net.peer.ip"),
                          _blocking.ffi.new("char[]", f"{ip_address}".encode('ascii'))))



        # add url as part of header processing
        pairs.append((_blocking.ffi.new("char[]", b"http.url"),
                      _blocking.ffi.new("char[]", f"{url}".encode('ascii'))))

        for header_key, header_value in headers.items():
            pairs.append((_blocking.ffi.new("char[]", f"{HTTP_HEADER_PREFIX}.{header_key}".lower().encode('ascii')),
                          _blocking.ffi.new("char[]", f"{header_value}".encode('ascii'))))

        num_headers = len(pairs)
        attributes = _blocking.ffi.new("traceable_attributes*")
        attribute_array = _blocking.ffi.new("traceable_attribute[]", num_headers)
        attributes.count = num_headers

        for index, pair in enumerate(pairs, start=0):
            attribute_array[index].key = pair[0]
            attribute_array[index].value = pair[1]
        attributes.attribute_array = attribute_array

        return self._evaluate_attributes(span, attributes)

    def evaluate_body(self, span: Span, body) -> bool:
        if not self.blocking_available:
            return False

        if not body or len(str(body)) == 0:
            return False

        attributes = _blocking.ffi.new("traceable_attributes*")
        attribute_array = _blocking.ffi.new("traceable_attribute[]", 1)
        attributes.count = 1
        attribute_array[0].key = _blocking.ffi.new("char[]", b"http.request.body")
        attribute_array[0].value = _blocking.ffi.new("char[]", f"{body}".encode('ascii'))

        attributes.attribute_array = attribute_array

        return self._evaluate_attributes(span, attributes)

    def _evaluate_attributes(self, span, attributes) -> bool:
        blocking_result = _blocking.ffi.new("traceable_block_result*")
        result_code = _blocking.lib.traceable_block_request(self.blocking_engine, attributes[0], blocking_result)
        if result_code != _blocking.lib.TRACEABLE_SUCCESS:
            logger.debug ("traceable_block_request fail!")
            return False

        _add_span_attributes(span, blocking_result)

        _result_code = _blocking.lib.traceable_delete_block_result_data(blocking_result[0])
        if blocking_result.block:
            return True
        return False


def _add_span_attributes(span: Span, blocking_result):
    attr_count = blocking_result.attributes.count
    for i in range(attr_count):
        key_attr = blocking_result.attributes.attribute_array[i].key
        value_attr = blocking_result.attributes.attribute_array[i].value
        if key_attr == _blocking.ffi.NULL or value_attr == _blocking.ffi.NULL:
            continue

        key = _blocking.ffi.string(key_attr)
        value = _blocking.ffi.string(value_attr)

        span.set_attribute(key.decode('utf-8', 'backslashescape'), value.decode('utf-8', 'backslashescape'))


def _create_blocking_config():
    traceable_config = Config().config
    should_debug = 1 if os.environ.get("TA_LOG_LEVEL", "").upper() == "DEBUG" else 0
    should_debug = 1 if should_debug or traceable_config.blocking_config.debug_log else 0

    blocking_config = _blocking.ffi.new("traceable_blocking_config*")

    blocking_config.log_config.mode = should_debug

    traceable_opa = traceable_config.opa
    opa_server_url = _blocking.ffi.new("char[]", f"{traceable_opa.endpoint.value}".encode('ascii'))

    blocking_config.opa_config.opa_server_url = opa_server_url
    blocking_config.opa_config.log_to_console = 1
    blocking_config.opa_config.skip_verify = 0
    blocking_config.opa_config.debug_log = should_debug
    blocking_config.opa_config.min_delay = traceable_opa.poll_period_seconds.value
    blocking_config.opa_config.max_delay = traceable_opa.poll_period_seconds.value
    blocking_config.opa_config.logging_dir = _blocking.ffi.new("char[]", "".encode('ascii'))
    blocking_config.opa_config.logging_file_prefix = _blocking.ffi.new("char[]", "".encode('ascii'))

    modsecurity_enabled = 1 if traceable_config.blocking_config.modsecurity.enabled.value is True else 0
    blocking_config.modsecurity_config.enabled = modsecurity_enabled

    rb_enabled = 1 if traceable_config.blocking_config.region_blocking.enabled else 0
    blocking_config.rb_config.enabled = rb_enabled

    remote_config_enabled = 1 if traceable_config.blocking_config.remote_config.enabled.value is True else 0
    blocking_config.remote_config.enabled = remote_config_enabled

    remote_config_endpoint = traceable_config.blocking_config.remote_config.endpoint.value
    remote_url = _blocking.ffi.new("char[]", f"{remote_config_endpoint}".encode('ascii'))
    blocking_config.remote_config.remote_endpoint = remote_url

    remote_poll_period_seconds = traceable_config.blocking_config.remote_config.poll_period_seconds.value
    blocking_config.remote_config.poll_period_sec = remote_poll_period_seconds

    blocking_config.evaluate_body = 1 if traceable_config.blocking_config.evaluate_body.value is True else 0
    return blocking_config



DEFAULT = {
    'opa': {
        'enabled': True,
        'endpoint': 'http://localhost:8181',
        'poll_period_seconds': 30
    },
    'blocking_config': {
        'enabled': False,
        'debug_log': False,
        'evaluate_body': True,
        'modsecurity': {
            'enabled': True
        },
        'regionBlocking': {
            'enabled': True
        },
        'remote_config': {
            'enabled': True,
            'endpoint': 'localhost:5441',
            'poll_period_seconds': 30
        }
    }
}



import distro

# These imports have to be first to respect TA_ env vars in hypertrace agent
from hypertrace import env_var_settings
env_var_settings.ENV_VAR_PREFIXES.append('TA')

from hypertrace import agent  # pylint:disable=C0413
from hypertrace.agent.filter.registry import Registry # pylint:disable=C0413
from hypertrace.agent.custom_logger import get_custom_logger # pylint:disable=C0413
from hypertrace.agent import constants # pylint:disable=C0413

from traceableai.filter.traceable import Traceable, _BLOCKING_AVAILABLE  # pylint:disable=C0413
from traceableai.version import __version__ # pylint:disable=C0413

logger = get_custom_logger(__name__)



# We need to override the version that is reported to the platform from hypertrace
constants.TELEMETRY_SDK_VERSION = __version__

class Agent(agent.Agent):
    def __init__(self):
        super().__init__()
        logger.debug("Platform: %s", distro.id())
        logger.debug("Platform version: %s", distro.version())
        logger.debug('TraceableAI Agent version: %s', __version__)
        logger.debug("successfully initialized traceableai agent")


    def add_traceable_filter(self):  #pylint:disable=R0201
        logger.debug("in add_traceable_filter")
        if distro.id() == 'alpine':
            logger.info('blocking unavailable on alpine linux, skipping filter registration')
            return
        if not _BLOCKING_AVAILABLE:
            logger.info("blocking unavailable, skipping filter registration")
            return
        try:
            Registry().register(Traceable)
            logger.debug("successfully initialized traceable filter")
        except: # pylint:disable=W0702
            logger.info("failed to register traceable blocking filter")



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: traceable_config.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import wrappers_pb2 as google_dot_protobuf_dot_wrappers__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='traceable_config.proto',
  package='ai.traceable.agent.config.v1',
  syntax='proto3',
  serialized_options=b'Z-github.com/Traceableai/agent-config/gen/go/v1',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x16traceable_config.proto\x12\x1c\x61i.traceable.agent.config.v1\x1a\x1egoogle/protobuf/wrappers.proto\"\x84\x01\n\x0b\x41gentConfig\x12.\n\x03opa\x18\x01 \x01(\x0b\x32!.ai.traceable.agent.config.v1.Opa\x12\x45\n\x0f\x62locking_config\x18\x02 \x01(\x0b\x32,.ai.traceable.agent.config.v1.BlockingConfig\"\x9c\x01\n\x03Opa\x12+\n\x07\x65nabled\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.BoolValue\x12.\n\x08\x65ndpoint\x18\x02 \x01(\x0b\x32\x1c.google.protobuf.StringValue\x12\x38\n\x13poll_period_seconds\x18\x03 \x01(\x0b\x32\x1b.google.protobuf.Int32Value\"\xf5\x02\n\x0e\x42lockingConfig\x12+\n\x07\x65nabled\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.BoolValue\x12-\n\tdebug_log\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.BoolValue\x12\x44\n\x0bmodsecurity\x18\x03 \x01(\x0b\x32/.ai.traceable.agent.config.v1.ModsecurityConfig\x12\x31\n\revaluate_body\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.BoolValue\x12K\n\x0fregion_blocking\x18\x05 \x01(\x0b\x32\x32.ai.traceable.agent.config.v1.RegionBlockingConfig\x12\x41\n\rremote_config\x18\x06 \x01(\x0b\x32*.ai.traceable.agent.config.v1.RemoteConfig\"@\n\x11ModsecurityConfig\x12+\n\x07\x65nabled\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.BoolValue\"C\n\x14RegionBlockingConfig\x12+\n\x07\x65nabled\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.BoolValue\"\xa5\x01\n\x0cRemoteConfig\x12+\n\x07\x65nabled\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.BoolValue\x12.\n\x08\x65ndpoint\x18\x02 \x01(\x0b\x32\x1c.google.protobuf.StringValue\x12\x38\n\x13poll_period_seconds\x18\x03 \x01(\x0b\x32\x1b.google.protobuf.Int32ValueB/Z-github.com/Traceableai/agent-config/gen/go/v1b\x06proto3'
  ,
  dependencies=[google_dot_protobuf_dot_wrappers__pb2.DESCRIPTOR,])




_AGENTCONFIG = _descriptor.Descriptor(
  name='AgentConfig',
  full_name='ai.traceable.agent.config.v1.AgentConfig',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='opa', full_name='ai.traceable.agent.config.v1.AgentConfig.opa', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='blocking_config', full_name='ai.traceable.agent.config.v1.AgentConfig.blocking_config', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=89,
  serialized_end=221,
)


_OPA = _descriptor.Descriptor(
  name='Opa',
  full_name='ai.traceable.agent.config.v1.Opa',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='enabled', full_name='ai.traceable.agent.config.v1.Opa.enabled', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='endpoint', full_name='ai.traceable.agent.config.v1.Opa.endpoint', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='poll_period_seconds', full_name='ai.traceable.agent.config.v1.Opa.poll_period_seconds', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=224,
  serialized_end=380,
)


_BLOCKINGCONFIG = _descriptor.Descriptor(
  name='BlockingConfig',
  full_name='ai.traceable.agent.config.v1.BlockingConfig',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='enabled', full_name='ai.traceable.agent.config.v1.BlockingConfig.enabled', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='debug_log', full_name='ai.traceable.agent.config.v1.BlockingConfig.debug_log', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='modsecurity', full_name='ai.traceable.agent.config.v1.BlockingConfig.modsecurity', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='evaluate_body', full_name='ai.traceable.agent.config.v1.BlockingConfig.evaluate_body', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='region_blocking', full_name='ai.traceable.agent.config.v1.BlockingConfig.region_blocking', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='remote_config', full_name='ai.traceable.agent.config.v1.BlockingConfig.remote_config', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=383,
  serialized_end=756,
)


_MODSECURITYCONFIG = _descriptor.Descriptor(
  name='ModsecurityConfig',
  full_name='ai.traceable.agent.config.v1.ModsecurityConfig',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='enabled', full_name='ai.traceable.agent.config.v1.ModsecurityConfig.enabled', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=758,
  serialized_end=822,
)


_REGIONBLOCKINGCONFIG = _descriptor.Descriptor(
  name='RegionBlockingConfig',
  full_name='ai.traceable.agent.config.v1.RegionBlockingConfig',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='enabled', full_name='ai.traceable.agent.config.v1.RegionBlockingConfig.enabled', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=824,
  serialized_end=891,
)


_REMOTECONFIG = _descriptor.Descriptor(
  name='RemoteConfig',
  full_name='ai.traceable.agent.config.v1.RemoteConfig',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='enabled', full_name='ai.traceable.agent.config.v1.RemoteConfig.enabled', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='endpoint', full_name='ai.traceable.agent.config.v1.RemoteConfig.endpoint', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='poll_period_seconds', full_name='ai.traceable.agent.config.v1.RemoteConfig.poll_period_seconds', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=894,
  serialized_end=1059,
)

_AGENTCONFIG.fields_by_name['opa'].message_type = _OPA
_AGENTCONFIG.fields_by_name['blocking_config'].message_type = _BLOCKINGCONFIG
_OPA.fields_by_name['enabled'].message_type = google_dot_protobuf_dot_wrappers__pb2._BOOLVALUE
_OPA.fields_by_name['endpoint'].message_type = google_dot_protobuf_dot_wrappers__pb2._STRINGVALUE
_OPA.fields_by_name['poll_period_seconds'].message_type = google_dot_protobuf_dot_wrappers__pb2._INT32VALUE
_BLOCKINGCONFIG.fields_by_name['enabled'].message_type = google_dot_protobuf_dot_wrappers__pb2._BOOLVALUE
_BLOCKINGCONFIG.fields_by_name['debug_log'].message_type = google_dot_protobuf_dot_wrappers__pb2._BOOLVALUE
_BLOCKINGCONFIG.fields_by_name['modsecurity'].message_type = _MODSECURITYCONFIG
_BLOCKINGCONFIG.fields_by_name['evaluate_body'].message_type = google_dot_protobuf_dot_wrappers__pb2._BOOLVALUE
_BLOCKINGCONFIG.fields_by_name['region_blocking'].message_type = _REGIONBLOCKINGCONFIG
_BLOCKINGCONFIG.fields_by_name['remote_config'].message_type = _REMOTECONFIG
_MODSECURITYCONFIG.fields_by_name['enabled'].message_type = google_dot_protobuf_dot_wrappers__pb2._BOOLVALUE
_REGIONBLOCKINGCONFIG.fields_by_name['enabled'].message_type = google_dot_protobuf_dot_wrappers__pb2._BOOLVALUE
_REMOTECONFIG.fields_by_name['enabled'].message_type = google_dot_protobuf_dot_wrappers__pb2._BOOLVALUE
_REMOTECONFIG.fields_by_name['endpoint'].message_type = google_dot_protobuf_dot_wrappers__pb2._STRINGVALUE
_REMOTECONFIG.fields_by_name['poll_period_seconds'].message_type = google_dot_protobuf_dot_wrappers__pb2._INT32VALUE
DESCRIPTOR.message_types_by_name['AgentConfig'] = _AGENTCONFIG
DESCRIPTOR.message_types_by_name['Opa'] = _OPA
DESCRIPTOR.message_types_by_name['BlockingConfig'] = _BLOCKINGCONFIG
DESCRIPTOR.message_types_by_name['ModsecurityConfig'] = _MODSECURITYCONFIG
DESCRIPTOR.message_types_by_name['RegionBlockingConfig'] = _REGIONBLOCKINGCONFIG
DESCRIPTOR.message_types_by_name['RemoteConfig'] = _REMOTECONFIG
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

AgentConfig = _reflection.GeneratedProtocolMessageType('AgentConfig', (_message.Message,), {
  'DESCRIPTOR' : _AGENTCONFIG,
  '__module__' : 'traceable_config_pb2'
  # @@protoc_insertion_point(class_scope:ai.traceable.agent.config.v1.AgentConfig)
  })
_sym_db.RegisterMessage(AgentConfig)

Opa = _reflection.GeneratedProtocolMessageType('Opa', (_message.Message,), {
  'DESCRIPTOR' : _OPA,
  '__module__' : 'traceable_config_pb2'
  # @@protoc_insertion_point(class_scope:ai.traceable.agent.config.v1.Opa)
  })
_sym_db.RegisterMessage(Opa)

BlockingConfig = _reflection.GeneratedProtocolMessageType('BlockingConfig', (_message.Message,), {
  'DESCRIPTOR' : _BLOCKINGCONFIG,
  '__module__' : 'traceable_config_pb2'
  # @@protoc_insertion_point(class_scope:ai.traceable.agent.config.v1.BlockingConfig)
  })
_sym_db.RegisterMessage(BlockingConfig)

ModsecurityConfig = _reflection.GeneratedProtocolMessageType('ModsecurityConfig', (_message.Message,), {
  'DESCRIPTOR' : _MODSECURITYCONFIG,
  '__module__' : 'traceable_config_pb2'
  # @@protoc_insertion_point(class_scope:ai.traceable.agent.config.v1.ModsecurityConfig)
  })
_sym_db.RegisterMessage(ModsecurityConfig)

RegionBlockingConfig = _reflection.GeneratedProtocolMessageType('RegionBlockingConfig', (_message.Message,), {
  'DESCRIPTOR' : _REGIONBLOCKINGCONFIG,
  '__module__' : 'traceable_config_pb2'
  # @@protoc_insertion_point(class_scope:ai.traceable.agent.config.v1.RegionBlockingConfig)
  })
_sym_db.RegisterMessage(RegionBlockingConfig)

RemoteConfig = _reflection.GeneratedProtocolMessageType('RemoteConfig', (_message.Message,), {
  'DESCRIPTOR' : _REMOTECONFIG,
  '__module__' : 'traceable_config_pb2'
  # @@protoc_insertion_point(class_scope:ai.traceable.agent.config.v1.RemoteConfig)
  })
_sym_db.RegisterMessage(RemoteConfig)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



import os
import traceback
import zipfile
import io
import urllib.request

import distro

# List is based on this workflow:
# https://github.com/Traceableai/libtraceable/blob/main/.github/workflows/build.yaml#L14-L43
from traceableai.filter import build

supported_platform_version = {
    "debian": {
        "os": "ubuntu",
        "versions": {
            "10": "18.04"
        }
    },
    "ubuntu": {
        "os": "ubuntu",
        "versions": {
            "18.04": "18.04",
            "20.04": "20.04"
        }
    },
    "centos": {
        "os": "centos",
        "versions": {
            "7": "7",
            "8": "8"
        }
    },
    # Disable alpine support for now
    # "alpine": {
    #     "os": "alpine",
    #     "versions": {
    #         "3.9": "3.9",
    #         "3.12": "3.12",
    #     }
    # },
    "amzn": {
        "os": "centos",
        "versions": {
            "2": "7"
        }
    }
}

LIBTRACEABLE_VERSION = "0.1.97"
SOURCE_BUILD_DIR = os.path.join(os.path.dirname(__file__), '..', 'filter')
DEFAULT_URL = "https://downloads.traceable.ai/install/libtraceable"
if 'TA_LIBTRACEABLE_URL' in os.environ:
    if len(os.environ['TA_LIBTRACEABLE_URL']) > 0:
        URL = os.environ['TA_LIBTRACEABLE_URL']
    else:
        URL = DEFAULT_URL
else:
    URL = DEFAULT_URL


def download_authenticated(platform, version, directory):
    libtraceable_name = f"libtraceable_{platform}_{version}_x86_64"
    filename = f"{libtraceable_name}-{LIBTRACEABLE_VERSION}.zip"
    url = f"{URL}/{libtraceable_name}/{LIBTRACEABLE_VERSION}/{filename}"

    username = os.environ["TA_BASIC_AUTH_USER"]
    password = os.environ["TA_BASIC_AUTH_TOKEN"]

    pass_manager = urllib.request.HTTPPasswordMgrWithDefaultRealm()
    pass_manager.add_password(None, url, username, password)
    authhandler = urllib.request.HTTPBasicAuthHandler(pass_manager)

    opener = urllib.request.build_opener(authhandler)
    urllib.request.install_opener(opener)

    response = urllib.request.urlopen(url)  # pylint:disable=R1732
    return _extract_library_files(response, directory)


def download(platform, version, directory):
    print('downloading from public repo')
    libtraceable_name = f"libtraceable_{platform}_{version}_x86_64"
    filename = f"{libtraceable_name}-{LIBTRACEABLE_VERSION}.zip"
    url = f"{URL}/{libtraceable_name}/{LIBTRACEABLE_VERSION}/{filename}"
    response = urllib.request.urlopen(url)  # pylint:disable=R1732
    return _extract_library_files(response, directory)


def install_lib_traceable(directory=SOURCE_BUILD_DIR):
    try:
        platform = distro.id()
        version = distro.version()
        print(f"Platform: {platform}")
        print(f"Version: {version}")

        support_dict = supported_platform_version.get(platform, None)
        if support_dict is None:
            print(f"libtraceable is not supported on this platform - {platform}-{version}")
            return

        download_version = supported_version(version, support_dict)
        if download_version is None:
            print(f"libtraceable is not supported on this version of this platform {platform}-{version}")
            return

        platform = support_dict["os"]
        print(f"Attempting to download libtraceable for {platform}_{download_version}")
        if "TA_BASIC_AUTH_USER" in os.environ and "TA_BASIC_AUTH_TOKEN" in os.environ:
            download_authenticated(platform, download_version, directory)
        else:
            download(platform, download_version, directory)
        build(directory)
        print("Done downloading - attempting build")
    except:  # pylint:disable=W0702
        print("Failed to download libtraceable %s", traceback.format_exc())
        # if we fail to install or build we don't want to break pip install


def supported_version(actual_version, support_dict):
    supported_versions = support_dict['versions']
    for version_key in supported_versions:
        if actual_version.startswith(version_key):
            return supported_versions[version_key]

    return supported_versions[list(supported_versions.keys())[0]]


def _extract_library_files(response, directory):
    print("extracting library files")
    with zipfile.ZipFile(io.BytesIO(response.read())) as archive:
        for zip_info in archive.infolist():
            if zip_info.filename[-1] == '/':
                continue
            print(f"extracted: {zip_info.filename}")
            zip_info.filename = os.path.basename(zip_info.filename)
            archive.extract(zip_info, directory)



# Allow use of TA_ env vars instead of HT_
from hypertrace import env_var_settings

from traceableai.agent import Agent

env_var_settings.ENV_VAR_PREFIXES.append('TA')

# This will initialize the hypertrace instrumentation
a = Agent()
import hypertrace.agent.autoinstrumentation.sitecustomize # pylint:disable=C0413,W0611,C0411


from traceableai.config.config import Config  # pylint:disable=C0413,C0412

from hypertrace.agent.custom_logger import get_custom_logger  # pylint:disable=C0413,C0412,C0411

config = Config()
logger = get_custom_logger(__name__)
if config.config.blocking_config.enabled.value is True:
    a.add_traceable_filter()
else:
    logger.info("Not adding traceableai blocking module")



# pylint: skip-file
import os
from setuptools import setup, find_packages, Extension

exec(open('src/traceableai/version.py').read())

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="traceable-agent",
    version=__version__,
    author="Traceable.ai",
    description="Traceable.ai Python Agent",
    url="https://traceable.ai",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent"
    ],
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    include_package_data=True,
    python_requires=">=3.6",
    ext_modules=[ # this option will let wheel builder know to package c extensions when ran from CI. No effect for source tar installs
        Extension(
            name='traceableai.filter._blocking',
            sources=[],
            optional=os.environ.get('CIBUILDWHEEL', '0') != '1'
        )
    ],
    install_requires=[
        "hypertrace-agent==0.7.2",
        "distro==1.6.0",
        "cffi",
    ],
    entry_points = {
        'console_scripts': [
            'traceableai-instrument = traceableai.autoinstrumentation.wrapper:run',
            'traceableai-install = traceableai.autoinstrumentation.download:install_lib_traceable'
        ],
    }
)



import copy
import json
import logging
import os
from os import environ

import yaml

from google.protobuf import json_format

from traceableai.config import traceable_config_pb2
from traceableai.config.default import DEFAULT
from traceableai.config.environment import overwrite_with_environment
logger = logging.getLogger(__name__)

class Config(): # pylint:disable=R0903
    def __new__(cls):
        if not hasattr(cls, 'instance'):
            cls.instance = super(Config, cls).__new__(cls)
        return cls.instance

    def __init__(self):
        if not hasattr(self, '_initialized'):
            self._initialized = True
            self.config = build_config()


# We don't want to update a value that is set with an unset value
def merge_config(base_config, overriding_config):
    for key in overriding_config:
        if key in base_config and isinstance(base_config[key], dict):
            if key in overriding_config:
                base_config[key] = merge_config(base_config[key], overriding_config[key])
        else:
            base_config[key] = overriding_config[key]
    return base_config
# Config load ordering
# 1.) Use the defaults
# 2.) Override the defaults with config file
# 3.) Override the config file with env vars
def build_config():
    traceable_config = traceable_config_pb2.AgentConfig()
    config_dict = copy.deepcopy(DEFAULT)
    file_dict = read_from_file()
    if file_dict is not None:
        merge_config(config_dict, file_dict)

    json_string = json.dumps(config_dict)
    json_format.Parse(json_string, traceable_config, ignore_unknown_fields=True)

    overwrite_with_environment(traceable_config)

    return traceable_config

def read_from_file():
    config_path = _config_file_path()
    if config_path is None or not os.path.exists(config_path):
        logger.debug("TA_CONFIG_FILE path not set")
        return None

    with open(config_path, 'r') as config_file:
        try:
            return yaml.safe_load(config_file)
        except yaml.YAMLError as exc:
            logger.debug(exc)
            return None

def _config_file_path():
    config_path = environ.get('TA_CONFIG_FILE', None)
    if config_path is None:
        return None
    return config_path



import os

def _is_true(env_key):
    value = os.environ.get(env_key)
    return value.lower() == 'true'


def overwrite_with_environment(config):
    if "TA_BLOCKING_CONFIG_ENABLED" in os.environ:
        config.blocking_config.enabled.value = _is_true("TA_BLOCKING_CONFIG_ENABLED")

    if "TA_BLOCKING_CONFIG_DEBUG_LOG" in os.environ:
        config.blocking_config.debug_log.value = _is_true("TA_BLOCKING_CONFIG_DEBUG_LOG")

    if "TA_BLOCKING_CONFIG_EVALUATE_BODY" in os.environ:
        config.blocking_config.evaluate_body.value = _is_true("TA_BLOCKING_CONFIG_EVALUATE_BODY")

    if "TA_BLOCKING_CONFIG_MODSECURITY_ENABLED" in os.environ:
        config.blocking_config.modsecurity.enabled.value = _is_true("TA_BLOCKING_CONFIG_MODSECURITY_ENABLED")

    if "TA_BLOCKING_CONFIG_REGION_BLOCKING_ENABLED" in os.environ:
        config.blocking_config.region_blocking.enabled.value = _is_true("TA_BLOCKING_CONFIG_REGION_BLOCKING_ENABLED")

    if "TA_BLOCKING_CONFIG_REMOTE_CONFIG_ENABLED" in os.environ:
        config.blocking_config.remote_config.enabled.value = _is_true("TA_BLOCKING_CONFIG_REMOTE_CONFIG_ENABLED")

    if "TA_BLOCKING_CONFIG_REMOTE_CONFIG_ENDPOINT" in os.environ:
        config.blocking_config.remote_config.endpoint.value = os.environ["TA_BLOCKING_CONFIG_REMOTE_CONFIG_ENDPOINT"]

    if "TA_BLOCKING_CONFIG_REMOTE_CONFIG_POLL_PERIOD_SECONDS" in os.environ:
        config.blocking_config.remote_config.poll_period_seconds.value = \
            int(os.environ["TA_BLOCKING_CONFIG_REMOTE_CONFIG_POLL_PERIOD_SECONDS"])

    if "TA_OPA_ENABLED" in os.environ:
        config.opa.enabled.value = _is_true("TA_OPA_ENABLED")

    if "TA_OPA_ENDPOINT" in os.environ:
        config.opa.endpoint.value = os.environ['TA_OPA_ENDPOINT']

    if "TA_OPA_POLL_PERIOD_SECONDS" in os.environ:
        config.opa.poll_period_seconds.value = int(os.environ['TA_OPA_POLL_PERIOD_SECONDS'])
